@@ -21,25 +21,30 @@
 */
 
 #include "dynamicqmetaobject.h"
+#include "dynamicqmetaobject_p.h"
+#include "pysidesignal.h"
+#include "pysidesignal_p.h"
+#include "pysideproperty.h"
+#include "pysideproperty_p.h"
+#include "pysideslot_p.h"
+
 #include <QByteArray>
 #include <QString>
 #include <QStringList>
 #include <QList>
+#include <QLinkedList>
 #include <QObject>
 #include <cstring>
 #include <QDebug>
 #include <QMetaMethod>
 
-#include "qsignal.h"
-#include "qproperty.h"
-
 #define MAX_SIGNALS_COUNT 50
 #define MAX_SLOTS_COUNT 50
 
 #define MAX_GLOBAL_SIGNALS_COUNT 500
 #define MAX_GLOBAL_SLOTS_COUNT 500
 
-#define GLOBAL_RECEIVER_CLASS_NAME "__GlobalReceiver__"
+#define EMPTY_META_METHOD "0()"
 
 using namespace PySide;
 
@@ -66,6 +71,18 @@ enum PropertyFlags  {
     Notify = 0x00400000
 };
 
+class DynamicQMetaObject::DynamicQMetaObjectPrivate
+{
+public:
+    QList<MethodData> m_signals;
+    QList<MethodData> m_slots;
+    QList<PropertyData> m_properties;
+    QByteArray m_className;
+
+    void updateMetaObject(QMetaObject* metaObj);
+    void writeMethodsData(const QList<MethodData>& methods, unsigned int** data, QList<QByteArray>* strings, int* prtIndex, int maxCount, int nullIndex, int flags);
+};
+
 static int registerString(const QByteArray& s, QList<QByteArray>* strings)
 {
     int idx = 0;
@@ -116,24 +133,17 @@ static bool isQRealType(const char *type)
 }
 
 
-
 /*
  * Avoid API break keep this on cpp
  */
-static int maxSlotsCount(const QString& className)
+static inline int maxSlotsCount(const QByteArray& className)
 {
-    int maxSlots = MAX_SLOTS_COUNT;
-    if (className == GLOBAL_RECEIVER_CLASS_NAME)
-        maxSlots = MAX_GLOBAL_SIGNALS_COUNT;
-    return maxSlots;
+    return className == GLOBAL_RECEIVER_CLASS_NAME ? MAX_GLOBAL_SIGNALS_COUNT : MAX_SLOTS_COUNT;
 }
 
-static int maxSignalsCount(const QString& className)
+static inline int maxSignalsCount(const QByteArray& className)
 {
-    int maxSignals = MAX_SIGNALS_COUNT;
-    if (className == GLOBAL_RECEIVER_CLASS_NAME)
-        maxSignals = MAX_GLOBAL_SIGNALS_COUNT;
-    return maxSignals;
+    return className == GLOBAL_RECEIVER_CLASS_NAME ? MAX_GLOBAL_SIGNALS_COUNT : MAX_SIGNALS_COUNT;
 }
 
 uint PropertyData::flags() const
@@ -145,103 +155,107 @@ uint PropertyData::flags() const
     else if (!isQRealType(typeName))
         flags |= qvariant_nameToType(typeName) << 24;
 
-    if (qproperty_is_readble(m_data))
+    if (PySide::Property::isReadable(m_data))
         flags |= Readable;
 
-    if (qproperty_is_writable(m_data))
+    if (PySide::Property::isWritable(m_data))
         flags |= Writable;
 
-    if (qproperty_has_reset(m_data))
+    if (PySide::Property::hasReset(m_data))
         flags |= Resettable;
 
-    if (!qproperty_is_designable(m_data))
-        flags |= ResolveDesignable;
-    else
+    if (PySide::Property::isDesignable(m_data))
         flags |= Designable;
-
-    if (!qproperty_is_scriptable(m_data))
-        flags |= ResolveScriptable;
     else
+        flags |= ResolveDesignable;
+
+    if (PySide::Property::isScriptable(m_data))
         flags |= Scriptable;
-
-    if (!qproperty_is_stored(m_data))
-        flags |= ResolveStored;
     else
+        flags |= ResolveScriptable;
+
+    if (PySide::Property::isStored(m_data))
         flags |= Stored;
-
-    if (!qproperty_is_user(m_data))
-        flags |= ResolveUser;
     else
-        flags |= User;
+        flags |= ResolveStored;
 
-    if (qproperty_is_constant(m_data))
+    //EDITABLE
+    flags |= ResolveEditable;
+
+    if (PySide::Property::isUser(m_data))
+        flags |= User;
+    else
+        flags |= ResolveUser;
+
+    if (m_notifyId != -1)
+        flags |= Notify;
+
+    if (PySide::Property::isConstant(m_data))
         flags |= Constant;
 
-    if (qproperty_is_final(m_data))
+    if (PySide::Property::isFinal(m_data))
         flags |= Final;
 
     return flags;
 }
 
-MethodData::MethodData(const char* signature, const char* type)
+// const QByteArray with EMPTY_META_METHOD, used to save some memory
+const QByteArray MethodData::m_emptySig(EMPTY_META_METHOD);
+
+MethodData::MethodData() : m_signature(m_emptySig)
+{
+}
+
+MethodData::MethodData(const char* signature, const char* type) : m_signature(signature), m_type(type)
 {
-    m_signature = QSharedPointer<QByteArray>(new QByteArray(signature));
-    m_type = QSharedPointer<QByteArray>(new QByteArray(type));
 }
 
 void MethodData::clear()
 {
-    m_signature->clear();
-    m_type->clear();
+    m_signature = m_emptySig;
+    m_type.clear();
 }
 
 bool MethodData::operator==(const MethodData& other) const
 {
-    return *m_signature == other.signature();
+    return m_signature == other.signature();
 }
 
 bool MethodData::operator==(const char* other) const
 {
-    return *m_signature == other;
+    return m_signature == other;
 }
 
 QByteArray MethodData::signature() const
 {
-    if (!m_signature.isNull())
-        return *m_signature;
-    else
-        return QByteArray();
+    return m_signature;
 }
 
 QByteArray MethodData::type() const
 {
-    if (!m_type.isNull()) {
-        if (*m_type == "void")
-            return QByteArray();
-        return *m_type;
-    } else {
+    if (m_type == "void")
         return QByteArray();
-    }
+    return m_type;
 }
 
 bool MethodData::isValid() const
 {
-    return m_signature->size();
+    return m_signature.size();
 }
 
 PropertyData::PropertyData()
-    : m_data(0)
+    : m_notifyId(0), m_data(0)
 {
 }
 
-PropertyData::PropertyData(const char* name, PyObject* data)
-    : m_name(name), m_data(data)
+PropertyData::PropertyData(const char* name, int notifyId, PySideProperty* data)
+    : m_name(name), m_notifyId(notifyId), m_data(data)
 {
 }
 
 QByteArray PropertyData::type() const
 {
-    return QByteArray(qproperty_get_type(m_data));
+    return QByteArray(PySide::Property::getTypeName(m_data));
 }
 
 
@@ -255,6 +269,11 @@ QByteArray PropertyData::name() const
     return m_name;
 }
 
+int PropertyData::notifyId() const
+{
+    return m_notifyId;
+}
+
 bool PropertyData::operator==(const PropertyData& other) const
 {
     return m_data == other.m_data;
@@ -266,192 +285,158 @@ bool PropertyData::operator==(const char* name) const
 }
 
 
-DynamicQMetaObject::DynamicQMetaObject(const char* className, const QMetaObject* metaObject)
+DynamicQMetaObject::DynamicQMetaObject(const char* className, const QMetaObject* metaObject) : m_d(new DynamicQMetaObjectPrivate)
 {
     d.superdata = metaObject;
     d.stringdata = 0;
     d.data = 0;
     d.extradata = 0;
-    m_className = QByteArray(className);
-    updateMetaObject();
+    m_d->m_className = className;
+    m_d->updateMetaObject(this);
 }
 
 DynamicQMetaObject::~DynamicQMetaObject()
 {
     delete[] d.stringdata;
     delete[] d.data;
+    delete m_d;
 }
 
 void DynamicQMetaObject::addSignal(const char* signal, const char* type)
 {
-    QLinkedList<MethodData>::iterator i = qFind(m_signals.begin(), m_signals.end(), signal);
-    if (i != m_signals.end())
+    int index = m_d->m_signals.indexOf(signal);
+    if (index != -1)
         return;
 
     //search for a empty space
     MethodData blank;
-    i = qFind(m_signals.begin(), m_signals.end(), blank);
-    if (i != m_signals.end()) {
-        *i = MethodData(signal, type);
-        updateMetaObject();
+    index = m_d->m_signals.indexOf(blank);
+    if (index != -1) {
+        m_d->m_signals[index] = MethodData(signal, type);
+        m_d->updateMetaObject(this);
         return;
     }
 
-    int maxSignals = maxSignalsCount(m_className);
-    if (m_signals.size() >= maxSignals) {
+    int maxSignals = maxSignalsCount(m_d->m_className);
+    if (m_d->m_signals.size() >= maxSignals) {
         qWarning() << "Fail to add dynamic signal to QObject. PySide support at most" << maxSignals << "dynamic signals.";
         return;
     }
 
-    m_signals << MethodData(signal, type);
-    updateMetaObject();
+    m_d->m_signals << MethodData(signal, type);
+    m_d->updateMetaObject(this);
 }
 
 void DynamicQMetaObject::addSlot(const char* slot, const char* type)
 {
-    QLinkedList<MethodData>::iterator i = qFind(m_slots.begin(), m_slots.end(), slot);
-    if (i != m_slots.end())
+    int index = m_d->m_slots.indexOf(slot);
+    if (index != -1)
         return;
 
-    int maxSlots = maxSlotsCount(m_className);
-    if (m_slots.size() >= maxSlots) {
-        qWarning() << "Fail to add dynamic slot to QObject. PySide support at most" << maxSlots << "dynamic slots.";
-        return;
-    }
-
     //search for a empty space
     MethodData blank;
-    i = qFind(m_slots.begin(), m_slots.end(), blank);
-    if (i != m_slots.end())
-        *i = MethodData(slot, type);
-    else
-        m_slots << MethodData(slot, type);
-    updateMetaObject();
+    index = m_d->m_slots.indexOf(blank);
+    if (index != -1) {
+        m_d->m_slots[index] = MethodData(slot, type);
+    } else if (m_d->m_slots.size() < maxSlotsCount(m_d->m_className)) {
+        m_d->m_slots << MethodData(slot, type);
+    } else {
+        qWarning() << "Fail to add dynamic slot to QObject. PySide support at most" << maxSlotsCount(m_d->m_className) << "dynamic slots.";
+        return;
+    }
+    m_d->updateMetaObject(this);
 }
 
 void DynamicQMetaObject::removeSlot(uint index)
 {
-    QMetaMethod m = method(index);
-    foreach(MethodData md, m_slots) {
-        if (md.signature() == m.signature()) {
-            md.clear();
-            updateMetaObject();
+    const char* methodSig = method(index).signature();
+    QList<MethodData>::iterator it = m_d->m_slots.begin();
+    for (; it != m_d->m_slots.end(); ++it) {
+        if (it->signature() == methodSig) {
+            it->clear();
+            m_d->updateMetaObject(this);
             break;
         }
     }
 }
 
-void DynamicQMetaObject::addProperty(const char* property, PyObject* data)
+void DynamicQMetaObject::addProperty(const char* propertyName, PyObject* data)
 {
-    QLinkedList<PropertyData>::iterator i = qFind(m_properties.begin(), m_properties.end(), property);
-    if (i != m_properties.end())
+    int index = m_d->m_properties.indexOf(propertyName);
+    if (index != -1)
         return;
 
+    // retrieve notifyId
+    int notifyId = -1;
+    PySideProperty* property = reinterpret_cast<PySideProperty*>(data);
+    if (property->d->notify) {
+        const char* signalNotify = PySide::Property::getNotifyName(property);
+        if (signalNotify) {
+            QByteArray signalSignature(signalNotify);
+            notifyId = m_d->m_signals.indexOf(signalNotify);
+        }
+    }
+
     //search for a empty space
     PropertyData blank;
-    i = qFind(m_properties.begin(), m_properties.end(), blank);
-    if (i != m_properties.end()) {
-        *i = PropertyData(property, data);
+    index = m_d->m_properties.indexOf(blank);
+    if (index != -1) {
+        m_d->m_properties[index] = PropertyData(propertyName, notifyId, property);
     } else {
-        m_properties << PropertyData(property, data);
+        m_d->m_properties << PropertyData(propertyName, notifyId, property);
     }
-    updateMetaObject();
-}
-
-
-DynamicQMetaObject* DynamicQMetaObject::createBasedOn(PyObject* pyObj, PyTypeObject* type, const QMetaObject* base)
-{
-    PyObject* key;
-    PyObject* value;
-    Py_ssize_t pos = 0;
-
-    QString className(type->tp_name);
-    className = className.mid(className.lastIndexOf('.')+1);
-    DynamicQMetaObject *mo = new PySide::DynamicQMetaObject(className.toAscii(), base);
-
-    while (PyDict_Next(type->tp_dict, &pos, &key, &value)) {
-
-        //Register properties
-        if (value->ob_type == &QProperty_Type) {
-            mo->addProperty(PyString_AsString(key), value);
-        }
-
-        //Register signals
-        if (value->ob_type == &Signal_Type) {
-            PyObject *attr = PyObject_GetAttr(pyObj, key);
-            SignalInstanceData *data = reinterpret_cast<SignalInstanceData*>(attr);
-            while(data) {
-                int index = base->indexOfSignal(data->signature);
-                if (index == -1)
-                    mo->addSignal(data->signature);
-                data = reinterpret_cast<SignalInstanceData*>(data->next);
-            }
-        }
-
-        if (!PyFunction_Check(value))
-            continue;
-
-        //Register Slots
-        if (PyObject_HasAttrString(value, PYSIDE_SLOT_LIST_ATTR)) {
-            PyObject *signature_list = PyObject_GetAttrString(value, PYSIDE_SLOT_LIST_ATTR);
-            for(Py_ssize_t i = 0, i_max = PyList_Size(signature_list); i < i_max; i++) {
-                PyObject *signature = PyList_GET_ITEM(signature_list, i);
-                QString sig(PyString_AsString(signature));
-                //slot the slot type and signature
-                QStringList slotInfo = sig.split(" ", QString::SkipEmptyParts);
-                int index = base->indexOfSlot(slotInfo[0].toAscii());
-                if (index == -1)
-                    mo->addSlot(slotInfo[1].toAscii(), slotInfo[0].toAscii());
-            }
-        }
-    }
-    return mo;
+    m_d->updateMetaObject(this);
 }
 
 void DynamicQMetaObject::removeSignal(uint index)
 {
     //Current Qt implementation does not support runtime remove signal
     QMetaMethod m = method(index);
-    foreach(MethodData md, m_signals) {
+    foreach(MethodData md, m_d->m_signals) {
         if (md.signature() == m.signature()) {
             md.clear();
-            updateMetaObject();
+            m_d->updateMetaObject(this);
             break;
         }
     }
 }
 
-void DynamicQMetaObject::writeMethodsData(QLinkedList<MethodData>& methods,
-                                          unsigned int **data,
-                                          QList<QByteArray> *strings,
-                                          int *prtIndex,
-                                          int max_count,
-                                          int null_index,
-                                          int flags)
+void DynamicQMetaObject::DynamicQMetaObjectPrivate::writeMethodsData(const QList<MethodData>& methods,
+                                                                     unsigned int** data,
+                                                                     QList<QByteArray>* strings,
+                                                                     int* prtIndex,
+                                                                     int maxCount,
+                                                                     int nullIndex,
+                                                                     int flags)
 {
     int index = *prtIndex;
 
-    QLinkedList<MethodData>::iterator iMethod = methods.begin();
-    for(int i=0; i < max_count; i++) {
-        QByteArray mType;
-        if (iMethod != methods.end() && ((*iMethod).signature().size() > 0) ) {
-            (*data)[index++] = registerString((*iMethod).signature(), strings); // func name
-            mType = (*iMethod).type();
-        } else {
-            (*data)[index++] = null_index; // func name
-        }
-        (*data)[index++] = null_index; // arguments
-        (*data)[index++] = (mType.size() > 0 ? registerString(mType, strings) : null_index); // normalized type
-        (*data)[index++] = null_index; // tags
+    int emptyIndex = registerString(EMPTY_META_METHOD, strings);
+
+    QList<MethodData>::const_iterator it = methods.begin();
+
+    for (; it != methods.end(); ++it) {
+        if (it->signature() != EMPTY_META_METHOD)
+            (*data)[index++] = registerString(it->signature(), strings); // func name
+        else
+            (*data)[index++] = emptyIndex; // func name
+        (*data)[index++] = nullIndex; // arguments
+        (*data)[index++] = (it->type().size() > 0 ? registerString(it->type(), strings) : nullIndex); // normalized type
+        (*data)[index++] = nullIndex; // tags
         (*data)[index++] = flags;
-        if (iMethod != methods.end()) 
-            iMethod++;
     }
 
+    for (int i = methods.count(); i < maxCount; ++i) {
+        (*data)[index++] = emptyIndex; // func name
+        (*data)[index++] = nullIndex; // arguments
+        (*data)[index++] = nullIndex; // normalized type
+        (*data)[index++] = nullIndex; // tags
+        (*data)[index++] = flags;
+    }
     *prtIndex = index;
 }
 
-void DynamicQMetaObject::updateMetaObject()
+void DynamicQMetaObject::DynamicQMetaObjectPrivate::updateMetaObject(QMetaObject* metaObj)
 {
     // these values are from moc source code, generator.cpp:66
     enum MethodFlags {
@@ -480,14 +465,15 @@ void DynamicQMetaObject::updateMetaObject()
                     n_properties, 0, // prop count and prop indexes
                     0, 0,         // enum count and enum index
                     0, 0,         // constructors
+                    0,            // flags
                     n_signals};
 
     const int HEADER_LENGHT = sizeof(header)/sizeof(int);
     header[5] = HEADER_LENGHT;
     // header size + 5 indexes per method + an ending zero
-    delete[] d.data;
+    delete[] metaObj->d.data;
     unsigned int* data;
-    data = new unsigned int[HEADER_LENGHT + n_methods*5 + n_properties*3 + 1];
+    data = new unsigned int[HEADER_LENGHT + n_methods*5 + n_properties*4 + 1];
     std::memcpy(data, header, sizeof(header));
 
     QList<QByteArray> strings;
@@ -515,6 +501,10 @@ void DynamicQMetaObject::updateMetaObject()
         data[index++] = pp.flags();
     }
 
+    //write properties notify
+    foreach(PropertyData pp, m_properties)
+        data[index++] = pp.notifyId() >= 0 ? pp.notifyId() : 0; //signal notify index
+
     data[index++] = 0; // the end
 
     // create the m_metadata string
@@ -524,9 +514,9 @@ void DynamicQMetaObject::updateMetaObject()
         str.append(char(0));
     }
 
-    delete[] d.stringdata;
+    delete[] metaObj->d.stringdata;
     char* stringData = new char[str.count()];
     std::copy(str.begin(), str.end(), stringData);
-    d.data = data;
-    d.stringdata = stringData;
+    metaObj->d.data = data;
+    metaObj->d.stringdata = stringData;
 }
